%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% embed_cllik:
%
% Do embedding by pairwise composite likelihood.
%
% Input:
%
% - Y: the data, binary encoded as -1s and 1s
% - L: loading coefficients
% - rho_X: covariance matrix of latent variables
% - X_children: list of 2 containing the of latent children
% - H: bi-directed graph
% - sd_r_prior: prior standard deviation for edge-copula parameters
% - X_v: grid values shared by all latent variables
% - X_ws: cell structure containing latent weights for a range of possible 
%         correlations
% - rho_domain: possible correlation values
% - M: number of grid points in the approximation
%
% Output:
% 
% - X_hat: estimated latent variables
%
% Created by: Ricardo Silva, London, 27/03/2012
% University College London
% Current version: 27/03/2012

function X_hat = embed_cllik(Y, L, rho_X, X_children, H, cop_hat)

% Initialization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Y_discard = abs(L(:, 1)) > 10;
L(Y_discard, :) = 0;

[N num_y] = size(Y);
num_x = length(X_children);
Y_parents = zeros(num_y, 1);
for x = 1:num_x
  for y = X_children{x} 
    Y_parents(y) = x;
  end
end

num_regions = length(cop_hat);
regions = zeros(2, num_regions); x_count = 1;
for x1 = 1:num_x    
  for x2 = (x1 + 1):num_x
    regions(:, x_count) = [x1; x2]; x_count = x_count + 1;
  end
end

options_optim = optimset('GradObj', 'on', 'LargeScale', 'on', ...
                         'Display', 'off', 'DerivativeCheck', 'off', ...
                         'TolFun', 1.e-3);

expo_y = get_expo_y(1:num_y, H);
X_hat = zeros(N, num_x);
X_start = init_embedding(L);

% Iterations %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for n = 1:N
  
  Y_n = (Y(n, :) + 1) / 2;
  x_0 = zeros(1, num_x);
  for x = 1:num_x
    y_count = 0;  
    for y = X_children{x}'
      if Y_discard(y), continue; end      
      x_0(x) = x_0(x) + X_start(y, Y_n(y) + 1);
      y_count = y_count + 1;
    end    
    x_0(x) = x_0(x) / y_count;
  end

  fprintf('> Optimizing: data point %d\n', n);
  
  X_hat(n, :) = fminunc(@(X)score_embedding(X, Y_n, L, rho_X, ...
                 Y_parents, X_children, H, regions, expo_y, cop_hat), ...
                 x_0, options_optim);
  
  save('embed_progress', 'X_hat', 'n');
end
  
% load('embed_progress'); scatter(X_hat(1:n, 1), X_hat(1:n, 3))
