%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initialize_pure_model:
%
% Initialize the parameters of a model with no bi-directed edges by fitting
% each one-factor component individually (assumes a minimum number of three
% indicators per component).
%
% Edge parameters and exponents are sampled using importance sampling.
%
% Input:
%
% - Y: the data, binary encoded as -1s and 1s
% - X_children: list of 2 containing the of latent children
% - X_v: grid values shared by all latent variables
% - X_ws: cell structure containing latent weights for a range of possible 
%         correlations
% - rho_domain: possible correlation values
%
% Output:
% 
% - L: loading matrix
% - rho_X: latent correlation (might not be positive definite)
%
% Created by: Ricardo Silva, London, 10/03/2012
% University College London
% Current version: 10/03/2012

function [L rho_X] = ...
    initialize_pure_model(Y, X_children, X_v, X_ws, rho_domain)

% Initialization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[N num_y] = size(Y);
num_x = length(X_children);
H = zeros(num_y);
Y_parents = zeros(num_y, 1);
for x = 1:num_x
  for y = X_children{x} 
    Y_parents(y) = x;
  end
end

num_regions = num_x * (num_x - 1) / 2;
regions = zeros(2, num_regions); x_count = 1;
for x1 = 1:num_x    
  for x2 = (x1 + 1):num_x
    regions(:, x_count) = [x1; x2]; x_count = x_count + 1;
  end
end

Y_exp_dummy = cell(num_regions, 1);
C_dummy = cell(num_regions, 1);
is_w_dummy = cell(num_regions, 1);
for r = 1:num_regions
  Y_exp_dummy{r} = eye(length(X_children{x}));
  C_dummy{r} = zeros(length(X_children{x}));
  is_w_dummy{r} = 1;
end

Y_counts = binary_sufficient(Y);

L = randn(num_y, 2);

options_optim = optimset('GradObj','on', 'LargeScale', 'on', 'Display', 'off');

% Fit loadings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for x = 1:num_x
  rho_X_dummy = [1 1; 1 1];
  theta_0 = reshape(L', 2 * num_y, 1);
  theta = fminunc(@(theta)score_parameters_cllik_mcdn(theta, rho_X_dummy, ...
          rho_domain, Y_counts, X_children, zeros(num_y), [x; x], ...
          Y_exp_dummy, C_dummy, is_w_dummy, X_v, X_ws, true, -N), ...
          theta_0, options_optim);
  L = reshape(theta, 2, num_y)';   
end

% Fit correlations %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rho_X = fit_rho_X_fixed_exp(L, Y_counts, X_children, H, regions, [], C_dummy, ...
                            is_w_dummy, X_v, X_ws, rho_domain, 0);
%rho_X = fit_rho_X(L, Y_counts, Y_parents, X_children, H, regions, ...
%                Y_exp_dummy, C_dummy, is_w_dummy, ...
%                X_v, X_ws, rho_domain);

