%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% laplace_appr_copula:
%
% Generate importance sample from the posterior distribution of a
% two-latent probit model. This does not treat the exponents as random.
%
% Input:
%
% - Y: the data, binary encoded as -1s and 1s
% - X_children: list of 2 containing the of latent children
% - H: bi-directed structure
% - L: prior variance for loadings
% - rho_X: latent correlation
% - expo_y: exponents used in each factor
% - sd_r_prior: prior standard deviation hyperparameters for the copula parameters
% - R0: starting point
% - M: number of points used in the "quadrature" approximation
%   
% Output:
% 
% - C, sample_w: representation of the posterior as a set of samples and weights
% - m, V: mean and covariance matrix of the Gaussian approximation
% - pairs: index of the vertices corresponding to each edge
%
% Created by: Ricardo Silva, London, 15/03/2012
% University College London
% Current version: 15/03/2012

function [C sample_w m V pairs new_rho_X] = ...
    laplace_appr_copula(Y, Y_parents, H, L, X_v, X_ws, rho_domain, ...
        expo_y, sd_r_prior, fix_rho, rho_X_fixed, R0, cutoff, M)
    
pairs = graph_edge_list(H); num_edges = size(pairs, 1);
if num_edges == 0
  C = []; sample_w = [];
  m = []; V = [];
  new_rho_X = rho_X_fixed;
  return
end

num_y = size(Y, 2);
S = count_types(Y);
all_Y = binary_comb(num_y);
sub_Y_0 = all_Y(S > 0, :);
S_sub = S(S > 0);         
num_sel = size(sub_Y_0, 1);

U = cdf_probit(L, Y_parents, X_v);

ws = zeros(num_edges, 2);
theta_0 = zeros(num_edges, 1);  
for i = 1:num_edges
  ws(i, :) = [expo_y(pairs(i, 1), pairs(i, 2)) ...
              expo_y(pairs(i, 2), pairs(i, 1))];
  theta_0(i) = R0(pairs(i, 1), pairs(i, 2));
end

[cliques, cliques_idx, ~, children, arg_message, mpo, ...
 clique_f, clique_u, clique_u_c] = message_passing_setup(H, pairs);

Y_partitioning = cell(length(cliques), 1);
for c = 1:length(cliques)
  Y_partitioning{c} = data_partitioning(sub_Y_0, cliques{c});
end

if ~fix_rho
  X_w_choice = X_ws;
else
  rho_idx = find(rho_domain == rho_X_fixed, 1);  
  X_w_choice = X_ws(:, rho_idx);
end   

options_optim = optimset('GradObj','off', 'LargeScale', 'off', 'Display', ...
                         'iter', 'DerivativeCheck', 'off');                     
m = fminunc(@(theta_raw)score_copula_mcdn(theta_raw), theta_0, options_optim);

V0 = reshape(hessian_dynare('score_copula_mcdn_dummy', m, 1.e-4, cutoff, ...
                    X_w_choice, U, pairs, ws, cliques, cliques_idx, ...
                    clique_f, clique_u, clique_u_c, expo_y, num_sel, ...
                    Y_partitioning, sub_Y_0, children, arg_message, mpo, ...
                    S_sub, sd_r_prior), length(m), length(m));
V = -inv(V0);
C = zeros(num_y, num_y, M);
sample_w = zeros(num_y, num_y, M);
for i = 1:num_edges
  if V(i, i) < 0    
    V(i, i) = 0.1; % Patch numerical instability
    fprintf('(* warning *) Laplace instability @edge [%d %d]\n', ...
            pairs(i, 1), pairs(i, 2));
  end 
  [points w] = build_distr_approx(M, [m(i) V(i, i) Inf]);  
  C(pairs(i, 1), pairs(i, 2), :) = points; 
  sample_w(pairs(i, 1), pairs(i, 2), :) = w;
  C(pairs(i, 2), pairs(i, 1), :) = C(pairs(i, 1), pairs(i, 2), :);
  sample_w(pairs(i, 2), pairs(i, 1), :) = sample_w(pairs(i, 1), pairs(i, 2), :);
end

  function score = score_copula_mcdn(theta_raw)
    theta = cop_transform(theta_raw, cutoff);  
    F = calc_copula_binary(U, pairs, ws, theta); 
    tables = tabulate(cliques, cliques_idx, clique_f, clique_u, clique_u_c, ...
                      U, diag(expo_y), pairs, F, num_sel, Y_partitioning, sub_Y_0);  
    pmf = cdf2pmf_mp(num_sel, X_w_choice, tables, cliques, cliques_idx, children, ...
                     arg_message, mpo);
    score = S_sub' * log(pmf) - 0.5 * (sum(theta_raw.^2) / sd_r_prior^2);
    if isnan(score)
      disp('!')
    end
    score = -score;% / N;
  end

end
