%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% learn_structure_mcdn_final:
%
% Learn structure of a measurement model with regions that include two 
% latent variables at a time only. Loadings and latent correlations are fit 
% by point estimation (and resulting estimate of latent correlations might
% not be positive definite). 
%
% The posterior distribution over edge parameters is approximated by the
% Laplace method. Exponents over arguments in each factor are generated
% deterministically according to the graph.
%
% Input:
%
% - Y: the data, binary encoded as -1s and 1s
% - Y_counts: bivariate sufficient statistics
% - X_children: list of 2 containing the of latent children
% - p_G: probability of anyedge existing
% - K: maximum number of optimization iterations
% - sd_r_prior: prior standard deviation for edge-copula parameters
% - X_v: grid values shared by all latent variables
% - X_ws: cell structure containing latent weights for a range of possible 
%         correlations
% - rho_domain: possible correlation values
% - M: number of discretization points
% - use_posterior: if true, use EM approach where posterior over copula
%                  parameters are derived for different two-latent regions
%
% Output:
% 
% - H: bi-directed structure
% - L: loading matrix
% - rho_X: latent correlation (might not be positive definite)
% - final_score: pairwise composite likelihood score
% - pair_scores: detailed scores per pair
% - num_iter: final number of iterations
%
% Created by: Ricardo Silva, London, 06/03/2012
% University College London
% Current version: 16/06/2012

function [H L rho_X final_score pair_scores num_iter] = ...
   learn_structure_mcdn_final(H_start, L_start, rho_X_start, ...
                              Y, Y_counts, X_children, p_G, K, ...
                              sd_r_prior, X_v, X_ws, rho_domain, M, ...
                              use_posterior)

% Initialization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[N num_y] = size(Y);
num_x = length(X_children);
Y_parents = zeros(num_y, 1);
for x = 1:num_x
  for y = X_children{x} 
    Y_parents(y) = x;
  end
end

L = L_start;
H = H_start;
rho_X = rho_X_start;

num_regions = num_x * (num_x - 1) / 2;
regions = zeros(2, num_regions); x_count = 1;
for x1 = 1:num_x    
  for x2 = (x1 + 1):num_x
    regions(:, x_count) = [x1; x2]; x_count = x_count + 1;
  end
end

[cop_theta_prior prior_w] = build_distr_approx(M, [0 sd_r_prior^2 Inf]);

options_optim = optimset('GradObj', 'off', 'LargeScale', 'off', ...
                         'Display', 'iter', 'DerivativeCheck', 'off', ...
                         'TolFun', 1.e-3);

if use_posterior
  C_region = cell(num_regions, 1);
  is_w_region = cell(num_regions, 1);

  cop_theta_start = cell(num_regions, 1); r_count = 1;
  for r = regions
    num_Y_scope = length(X_children{r(1)}) + length(X_children{r(2)});
    cop_theta_start{r_count} = reshape(sample_cop_theta_prior(num_Y_scope^2, ...
                               sd_r_prior, Inf), num_Y_scope, num_Y_scope);
    r_count = r_count + 1;
  end
else
  C_region = cell(num_regions, 1);
  is_w_region = cell(num_regions, 1);
  for i = 1:num_regions
    C_region{i} = zeros(num_y, num_y, M);
    is_w_region{i} = zeros(num_y, num_y, M);
    for j = 1:num_y
      for k = 1:num_y
        C_region{i}(j, k, :) = cop_theta_prior;
        is_w_region{i}(j, k, :) = prior_w;
      end
    end
  end    
end

expo_y = cell(num_regions, 1);
scale = 25;
fix_rho = true;

num_inner_i = 1;
num_iter = 1;


% Iterations %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for k = 1:K

  fprintf('> Optimizing: iteration %d\n', k);
  
  expo_y_all = get_expo_y(1:num_y, H);
  
  % Expectation step

  if use_posterior
    r_count = 1;
    for r = regions
      Y_scope = [X_children{r(1)}; X_children{r(2)}];
      Y_parents_r = [ones(length(X_children{r(1)}), 1); ...
                     2 * ones(length(X_children{r(2)}), 1)];    
      expo_y{r_count} = get_expo_y(Y_scope, H);
      H_r = H(Y_scope, Y_scope);
   
      [C_region{r_count}, is_w_region{r_count}, new_theta, ~, pairs] = ...
         laplace_appr_copula(Y(:, Y_scope), Y_parents_r, H_r, ...
            L(Y_scope, :), X_v, X_ws, rho_domain, expo_y{r_count}, ...
            sd_r_prior, fix_rho, rho_X(r(1), r(2)), cop_theta_start{r_count}, ... 
            scale, M);
      
      num_edges = size(pairs, 1);
      for i = 1:num_edges
        cop_theta_start{r_count}(pairs(i, 1), pairs(i, 2)) = new_theta(i);
        cop_theta_start{r_count}(pairs(i, 2), pairs(i, 1)) = new_theta(i);
      end
      r_count = r_count + 1;                                 
    end
  end

  % Maximization step: parameter point estimate

  rho_X = fit_rho_X_fixed_exp(L, Y_counts, ...
            X_children, H, regions, expo_y_all, C_region, is_w_region, ...
            X_v, X_ws, rho_domain, scale, zeros(num_x), zeros(num_x));
  for inner_i = 1:num_inner_i
    theta_0 = [L(:, 1); L(:, 2)]';
    theta = fminunc(@(theta)score_parameters_cllik_mcdn_fixed_exp(theta, ...
             rho_X, rho_domain, Y_counts, Y_parents, X_children, H, ...
              regions, expo_y_all, C_region, is_w_region, X_v, X_ws, true, ...
              scale, zeros(num_y, 1), zeros(num_y), -N), ...
                 theta_0, options_optim);
    L = reshape(theta, num_y, 2);
    rho_X = fit_rho_X_fixed_exp(L, Y_counts, ...
              X_children, H, regions, expo_y_all, C_region, is_w_region, ...
              X_v, X_ws, rho_domain, scale, zeros(num_x), zeros(num_x));
  end
  
  % Maximization step: adding/removing edges
  
  changed = false;
  
  while true
      
    best_1 = -1; best_2 = -1; best_score = -Inf;
    for y1 = 1:num_y
      for y2 = (y1 + 1):num_y
        old_edge = H(y1, y2);
        H(y1, y2) = 1; H(y2, y1) = 1;
        expo_y_all = get_expo_y(1:num_y, H);
        score1 = score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                    rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                    cop_theta_prior, prior_w, X_v, X_ws, fix_rho, scale, p_G);
        H(y1, y2) = 0; H(y2, y1) = 0;
        expo_y_all = get_expo_y(1:num_y, H);
        score0 = score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                    rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                    cop_theta_prior, prior_w, X_v, X_ws, fix_rho, scale, p_G); 
        H(y1, y2) = old_edge; H(y2, y1) = old_edge;
        if old_edge == 1
         score_change = score0 - score1;
         fprintf('> Change [%d %d DELETE] = %f ', y1, y2, score_change);
        else
         score_change = score1 - score0;
         fprintf('> Change [%d %d ADD] = %f ', y1, y2, score_change);
        end      
        if score_change > best_score && score_change > 0
          best_score = score_change;
          best_1 = y1; best_2 = y2;
          fprintf('(*)');
        end
        fprintf('\n');
      end
    end
  
    if best_score > 0
      changed = true;  
      if H(best_1, best_2) == 0
        fprintf('* Adding edge %d <--> %d\n', best_1, best_2);
        H(best_1, best_2) = 1; H(best_2, best_1) = 1;
      else
        fprintf('* Deleting edge %d <--> %d\n', best_1, best_2);
        H(best_1, best_2) = 0; H(best_2, best_1) = 0;  
      end      
      semi_score = 0; expo_y_all = get_expo_y(1:num_y, H);
      for y1 = 1:num_y
        for y2 = (y1 + 1):num_y  
          sub_score = ...
              score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                  rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                  cop_theta_prior, prior_w, X_v, X_ws, true, scale, p_G, false); 
          semi_score = semi_score + sub_score;
        end
      end
      semi_score = semi_score / N;
      fprintf('SEMI-SCORE = %f\n', semi_score);
    else
      break
    end
    
    break
    
  end
  
  if ~changed || k == K
    fprintf('FINISHED\n');
    final_score = 0; expo_y_all = get_expo_y(1:num_y, H);
    pair_scores = zeros(num_y);
    for y1 = 1:num_y
      for y2 = (y1 + 1):num_y  
        pair_scores(y1, y2) = ...
            score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                    rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                    cop_theta_prior, prior_w, X_v, X_ws, true, scale, p_G, false); 
        pair_scores(y2, y1) = pair_scores(y1, y2);
        final_score = final_score + pair_scores(y1, y2);            
      end
    end
    final_score = final_score / N;
    fprintf('Final score = %f\n', final_score);
    break
  end
  
  num_iter = k;
  %save('nhs_progress', 'H', 'L', 'rho_X', 'original_tabu_edges', 'tabu_values', 'num_iter');
end

