%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% tabulate:
%
% Generate tables for the "differential indicators" for all cliques 
% evaluates at data points.
%
% Input:
%
% - num_y: number of variables
% - cliques: list of cliques
% - U_y: unary values in observable space
% - expo_y: exponent of unvary factors
% - F_idx: list of (unary/pairwise) factor arguments in the observable
%          space
% - F_val: list of factor values (indexed by F_idx)
% - Y: evaluate table under these data points. THIS IS SUPPOSED TO BE
%      ENCODED AS 0\1 INSTEAD OF -1\1
% - grad_mode: if true, F(1, 1) = 0 instead of 1. This is because this
%              function is also used to calculate gradients. But only if
%              applied to a particular pair of variables {grad_1, grad_2}
%   
% Output:
% 
% - tables: cell structure with one table for each clique.
%
% Created by: Ricardo Silva, London, 22/02/2012
% University College London
% Current version: 22/02/2012

function tables = tabulate(cliques, cliques_idx, clique_f, clique_u, clique_u_c, ...
                           U_y, expo_y, F_idx, F_val, N_Y, Y_partitioning, ...
                           Y_alt, grad_mode, grad_1, grad_2)

num_cliques = length(cliques);
tables = cell(num_cliques, 1);

num_quad = size(F_val, 2);

U_yx = U_y'.^repmat(expo_y, 1, num_quad);

if nargin < 13
  grad_mode = false;
end

% Now for each clique populate table

for i = 1:num_cliques
    
  c_size = length(cliques{i});
  
  Z_delta = binary_comb(c_size);
  N_c = size(Z_delta, 1);
  tables{i} = cell(N_c, 1);
  for d = 1:N_c
    tables{i}{d} = zeros(N_Y, num_quad);
  end
  
  % Generate "marginal" CDFs as a function of *observable data Y*
  N_Y_c = N_c;       % Just an alias for readability
  Y_value = Z_delta; % Just an alias for readability
  F = ones(N_Y_c, num_quad); 
  for n = 1:N_Y_c
    for f = clique_f{i}
      y = cliques_idx(i, F_idx(f, :));
      if Y_value(n, y(1)) == 0 && Y_value(n, y(2)) == 0
        F(n, :) = F(n, :) .* F_val(f, :, 1);
      elseif Y_value(n, y(1)) == 1 && Y_value(n, y(2)) == 0
        F(n, :) = F(n, :) .* F_val(f, :, 2);
      elseif Y_value(n, y(1)) == 0 && Y_value(n, y(2)) == 1
        F(n, :) = F(n, :) .* F_val(f, :, 3);
      elseif grad_mode && ((F_idx(f, 1) == grad_1 && F_idx(f, 2) == grad_2) || (F_idx(f, 1) == grad_2 && F_idx(f, 2) == grad_1))
        F(n, :) = 0;
      end
    end
    for f_u = clique_u{i}
      y1 = cliques_idx(i, f_u);
      if Y_value(n, y1) == 0
        F(n, :) = F(n, :) .* U_yx(f_u, :);
      end
    end
  end
  
  % Generate table as a function of *difference data Z*  
  idx_w = 2.^((c_size - 1):-1:0)';  
  for d = 1:N_c
    sign = 1 - 2 * mod(sum(Z_delta(d, clique_u_c{i})), 2);  
    for n = 1:N_Y_c    
      Y_v = Y_value(n, :) - Z_delta(d, :);
      if any(Y_v < 0), continue; end
      idx_v = Y_partitioning{i}{n};
      idx_d = Y_v * idx_w + 1;
      tables{i}{d}(idx_v, :) = repmat(sign * F(idx_d, :), length(idx_v), 1);
    end
  end
  
  %table_alt = cell(N_c, 1);
  %for d = 1:N_c
  %  table_alt{d} = zeros(N_Y, num_quad);
  %  sign = 1 - 2 * mod(sum(Z_delta(d, clique_u_c{i})), 2);  
  %  
  %  Zs_delta = repmat(Z_delta(d, :), N_Y, 1);
  %  Y_v = Y_alt - Zs_delta;      
  %  sel = any(Y_v < 0, 2);
  %  
  %  idx_d = Y_v * idx_w + 1;
  %  table_alt{d}(sel == 0, :) = F(idx_d(sel == 0), :) * sign;  
  %end

end
