%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% calc_copula:
%
% Evaluates a set of pairwise copulas for binary data based on marginal 
% values U. Current this is for a Frank copula only.
%
% Input:
%
% - U: probability of variable being assigned its smaller value
% - pairs: indices of the respective variables in each copula
% - expo: exponents to which each element in each pair are raised
% - rho: parameters of each copula
%   
% Output:
% 
% - F: a table of copula evaluations. Each row corresponds to a different
%      copula, each column to a different combination of binary values 
%      (-1, -1), (-1, 1) and (1, -1) (values (1, 1) are redundant)
%
% Created by: Ricardo Silva, London, 21/02/2012
% University College London
% Current version: 21/02/2012

function F = calc_copula_binary(U, pairs, expo, rho)

num_x = size(U, 1);
num_pairs = size(pairs, 1);
F = zeros(num_pairs, num_x, 3);

values = [1 2 1; 1 1 2];
default_v = ones(num_pairs, 1);
e_rho = exp(-rho);

for x = 1:num_x
  v_count = 1;
  for v = values
    if v(1) == 1
      u_1 = U(x, pairs(:, 1))'.^expo(:, 1);
    else
      u_1 = default_v;
    end
    if v(2) == 1
      u_2 = U(x, pairs(:, 2))'.^expo(:, 2);
    else
      u_2 = default_v;
    end
    aux = 1 + (exp(-rho .* u_1) - 1) .* (exp(-rho .* u_2) - 1) ./ (e_rho - 1);
    F(:, x, v_count) = -log(aux) ./ rho;
    v_count = v_count + 1;
  end
end
