%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% cdf2pmf:
%
% Calculate the conditional probability mass function of a copula model 
% conditioned on covariates X, integrated away by a discretization.
%
% Input:
%
% - Y: data. THIS IS ASSUMED TO BE IN THE 0/1 FORMAT.
% - F_idx: indicates which variables appear in which F_val
% - F_val: pairwise CDF for each latent variable configuration and
%            observed variable configuration (3 of them, since the 4th is
%            redundant)
% - X_w: weights of the grid approximation for the latent variable
%        distribution
%
%   
% Output:
% 
% - pmf: conditional probability mass function for each combination of
%        binary instantations
%
% Created by: Ricardo Silva, London, 20/02/2012
% University College London
% Current version: 20/02/2012

function pmf = cdf2pmf(Y, U_y, expo_y, F_idx, F_val, X_w)

[N_Y num_y] = size(Y);

num_quad = size(F_val, 2);
num_factors = size(F_val, 1);

Y_delta = binary_comb(num_y);
N = size(Y_delta, 1);

U_yx = U_y'.^repmat(expo_y, 1, num_quad);

F = ones(N, num_quad);
for n = 1:N
  for f = 1:num_factors
    if Y_delta(n, F_idx(f, 1)) == 0 && Y_delta(n, F_idx(f, 2)) == 0
      F(n, :) = F(n, :) .* F_val(f, :, 1);
    elseif Y_delta(n, F_idx(f, 1)) == 1 && Y_delta(n, F_idx(f, 2)) == 0
      F(n, :) = F(n, :) .* F_val(f, :, 2);
    elseif Y_delta(n, F_idx(f, 1)) == 0 && Y_delta(n, F_idx(f, 2)) == 1
      F(n, :) = F(n, :) .* F_val(f, :, 3);
    end
  end
  for y = 1:num_y
    if Y_delta(n, y) == 0
      F(n, :) = F(n, :) .* U_yx(y, :);
    end
  end
end

pmf = zeros(N_Y, num_quad);
idx_w = 2.^((num_y - 1):-1:0)';
  
for d = 1:N    
  Ys_delta = repmat(Y_delta(d, :), N_Y, 1);
  Y_v = Y - Ys_delta;      
  sel = any(Y_v < 0, 2);
    
  idx_d = sum(Y_v * idx_w, 2) + 1;
  sign = 1 - 2 * mod(sum(Y_delta(d, :)), 2);
  pmf(sel == 0, :) = pmf(sel == 0, :) + F(idx_d(sel == 0), :) * sign;    
end
  
pmf = pmf * X_w;
