%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% message_passing_setup:
%
% Generates a list of cliques and junction tree for later message passing.
%
% Input:
%
% - G: the graph
% - F_idx: list of bivariate factors
%   
% Output:
% 
% - cliques: the vertices of the junction tree
% - cliques_idx: an index of the position of the vertices within each
%                clique ('0' where not present)
% - parents: the parent of each clique
% - children: the children of each clique
% - arg_message: for each tree vertex (clique), indicates what is the
%                domain of the message that needs to be passed to its 
%                parent
% - mpo: order in which messages should be calculated
% - clique_f: indicates, for each clique, which bivariate factors belong to it
% - clique_u: indicates, for each clique, which univariate factors belong to it
% - clique_u_c: same as above, but indexed wrt the relative position of
%               each variable within the clique
%
% Created by: Ricardo Silva, London, 22/02/2012
% University College London
% Current version: 29/02/2012

function [cliques cliques_idx parents children arg_message mpo...
          clique_f clique_u clique_u_c] = message_passing_setup(G, F_idx)

% First obtain junction tree

num_y = size(G, 1);
[~, cliques] = triangulate(G, 1:num_y);
%cliques = cell(1, 1); cliques{1} = 1:num_y; %TODO REMOVE THIS
num_cliques = length(cliques);
cliques_idx = zeros(num_cliques, num_y);
for i = 1:num_cliques
  y_pos = 1;
  for y = cliques{i}
    cliques_idx(i, y) = y_pos;
    y_pos = y_pos + 1;
  end
end

[jtree, root] = cliques_to_jtree(cliques, ones(num_y, 1) * 2);

parents = zeros(num_cliques, 1);
children = cell(num_cliques, 1);
arg_message = cell(num_cliques, 1);

% Generate order of messages and rearrangement of tree structure

marked = zeros(num_y, 1);
mpo = mpo_recurse(root);

% Now associate each factor with a single clique

num_factors = size(F_idx, 1);

clique_y_idx = zeros(num_y, num_cliques);
for i = 1:num_cliques
  clique_y_idx(cliques{i}, i) = 1;
end

clique_f = cell(num_cliques, 1);
for i = 1:num_factors
  y_1 = F_idx(i, 1); y_2 = F_idx(i, 2);
  choice_clique = find((clique_y_idx(y_1, :) .* clique_y_idx(y_2, :)) == 1);
  if length(choice_clique) == 1
    pick_clique = choice_clique;
  else
    pick_clique = randsample(choice_clique, 1);
  end
  clique_f{pick_clique} = [clique_f{pick_clique} i];
end

% Next associate each unary factor with a single clique

clique_u = cell(num_cliques, 1);
for y = 1:num_y
  choice_clique = find(clique_y_idx(y, :) == 1);
  if length(choice_clique) == 1
    pick_clique = choice_clique;
  else
    pick_clique = randsample(choice_clique, 1);
  end
  clique_u{pick_clique} = [clique_u{pick_clique} y];
end
clique_u_c = cell(num_cliques, 1);
for i = 1:num_cliques
  clique_u_c{i} = cliques_idx(i, clique_u{i});
end

  function order = mpo_recurse(start)
    marked(start) = 1;
    nn = find(jtree(start, :) == 1);
    nn = nn(marked(nn) == 0);
    children{start} = nn; 
    parents(nn) = start;
    branches = cell(length(nn));
    j = 1; length_order = 0;
    for n = nn      
      arg_message{n} = intersect(cliques{n}, cliques{start});
      branches{j} = mpo_recurse(n);
      length_order = length_order + length(branches{j});
      j = j + 1;
    end
    order = zeros(1, length_order + 1); 
    s = 1;
    for j = 1:length(nn)
      l = length(branches{j});
      order(s:(l + s - 1)) = branches{j};
      s = s + l;
    end
    order(end) = start; 
  end
  
end
